import sys
from pathlib import Path
import pytest

new_lib_path = str(Path(__file__).resolve().parents[1])

sys.path.append(new_lib_path)

from src.dmmp.exceptions import (
   MissingArgumentException,
   DirectoryAlreadyExists,
) 
                                 
import pytest

def test_missing_argument():
   with pytest.raises(MissingArgumentException):
      raise MissingArgumentException("test")


def test_directory_already_exists():
   with pytest.raises(DirectoryAlreadyExists):
      raise DirectoryAlreadyExists("test")


import tomllib
import tomli_w
import os
from .exceptions import MissingArgumentException


def _get_toml_args(
   path: str
   ):
   """Get the arguments in a .toml file
   
   :param path: The path of the .toml file
   """
   try:
      with open(path, "rb") as f:
         data = tomllib.load(f)
         return data
      
   except Exception as e:
      return {}


def get_args(
   config_file: str = None,
   dirs_to_map: list[tuple[str, str]] = None, 
   folders_ignore: list[str] = None,
   save_path: str = None,
   name_output: str = None,
   metadata_file_names: str = None
   ):
   """Merge the arguments passed by the CLI and the configuration file.

   Scan the configuration .toml file and the arguments passed in the
   CLI, creates a new dict and returns it, with the CLI arguments being
   the priority.

   :param config_file: .toml file to get configuration from.
   :param dirs_to_map: Directory to scan + number of nested folders to iterate.
   :param folders_ignore: List of folders to ignore.
   :param save_path: Path to save output.
   :param name_output: Name of the output directory.
   :param metadata_file_names: File names containing the directory metadata (no extension).   
   """

   # It's ok to be null, toml shouldnt be considered if not specified
   toml_dict = _get_toml_args(config_file)
   
   cli_dict = {
      k:v for k, v in {
      "config_file":config_file,
      "dirs_to_map":dirs_to_map,
      "folders_ignore":folders_ignore,
      "save_path":save_path,
      "name_output":name_output,
      "metadata_file_names":metadata_file_names,
      }.items() if v !=None
   }

   merged_dict = toml_dict | cli_dict

   # if there is no toml nor cli argument, fallback to default ones.
   # they weren't set as default arguments because the default arguments
   # would override the toml ones everytime, that's why they are done
   # after the merge
   merged_dict["folders_ignore"] = merged_dict.get("folders_ignore", [])
   merged_dict["save_path"] = merged_dict.get("save_path", os.getcwd())
   merged_dict["config_file"] = merged_dict.get("config_file", "config.toml")
   merged_dict["name_output"] = merged_dict.get("name_output", "dmmp-output") 
   merged_dict["metadata_file_names"] = \
      merged_dict.get("metadata_file_names", ["*"]) 

   return merged_dict


def validate_args(
   args: dict, 
   required_args: list[str]
   ):
   """Validate args.

   Compare if all args in a dict are present in required args.
   
   :param args: arguments.
   :param args: arguments that should be present in argument 
   keys.
   """
   for arg in required_args:
      if arg not in args.keys():
         raise MissingArgumentException(arg)
    

def write_args_to_toml(
   args: dict, 
   config_file_name: str
   ):
   """Write given args to a .toml file.

   :param args: arguments to write.
   :param config_file_name: the .toml output file name.
   """
   temp_args = args.copy()
   if temp_args.get("save_path") == os.getcwd():
      del temp_args["save_path"]

   if temp_args.get("config_file"):
      del temp_args["config_file"]
   
   with open(os.path.join(os.getcwd(), config_file_name), "wb") as f:
      tomli_w.dump(temp_args, f)


import sys
from pathlib import Path
import pytest

new_lib_path = str(Path(__file__).resolve().parents[1])

sys.path.append(new_lib_path)

from src.dmmp.args import _get_toml_args, validate_args, write_args_to_toml, get_args
from src.dmmp.exceptions import MissingArgumentException
import os
import json
import tomllib


generic_dict = {"data1":"1", "data2":"2"}

def test_get_toml_args():
   data = _get_toml_args(os.path.join(os.getcwd(), "tests", "test_content", "config.toml"))
   assert data == {"data":"test"}

def test_validate_args():
   validate_args(generic_dict, ["data1", "data2"])

   with pytest.raises(MissingArgumentException):
      validate_args({"not_data1":"1", "data2":"2"}, ["data1", "data2"])

def test_write_args_to_toml():
   path = os.path.join(os.getcwd(), "tests", "temp.toml")
   write_args_to_toml(generic_dict, path)
   data = _get_toml_args(path)
   assert data == generic_dict  
   os.remove(path)


def get_cli_args_and_output(key: str):
   cli_args = json.load(open("tests/cli_inputs.json"))[key]
   output = json.load(open("tests/outputs.json"))[key]
   return cli_args, output

def test_get_args_case1():
   cli_args = {
        "config_file": "test_content/case1.toml",
        "dirs_to_map": [["src", "2"], ["tests", "1"]],
        "folders_ignore": [".git", "__pycache__"],
        "save_path": "./output",
        "name_output": "scan-result",
        "metadata_file_names": ["meta", "info"]
   } 
   
   output = {
        "config_file": "test_content/case1.toml",
        "dirs_to_map": [["src", "2"], ["tests", "1"]],
        "folders_ignore": [".git", "__pycache__"],
        "save_path": "./output",
        "name_output": "scan-result",
        "metadata_file_names": ["meta", "info"]
   }
   assert(get_args(**cli_args) == output)
   

def test_get_args_case2():
   cli_args = {
        "dirs_to_map": [["src", "2"], ["tests", "1"]],
        "folders_ignore": [".git", "__pycache__"],
        "save_path": "./output",
        "name_output": "scan-result",
        "metadata_file_names": ["meta", "info"]
   } 
   
   output = {
        "config_file": "config.toml",
        "dirs_to_map": [["src", "2"], ["tests", "1"]],
        "folders_ignore": [".git", "__pycache__"],
        "save_path": "./output",
        "name_output": "scan-result",
        "metadata_file_names": ["meta", "info"]
   }
   assert(get_args(**cli_args) == output)

def test_get_args_case2():
   cli_args = {} 
   
   output = {
        "config_file": "config.toml",
        "folders_ignore": [],
        "save_path": os.getcwd(),
        "name_output": "dmmp-output",
        "metadata_file_names": ["*"]
   }
   assert(get_args(**cli_args) == output)

def test_get_args_case3():
   cli_args = {
      "config_file": "tests/test_content/case3.toml"
   } 
   
   output = {
        "config_file": "tests/test_content/case3.toml",
        "folders_ignore": [".venv", "node_modules"],
        "save_path": "C:/temp",
        "name_output": "specific-name",
        "metadata_file_names": ["desc", "test"]
   }
   assert(get_args(**cli_args) == output)



import os
from tqdm import tqdm
import re
from .exceptions import DirectoryAlreadyExists, DuplicateUniqueId, \
                           DuplicateOutputFile, NoRelationshipFound


class Mapper():
   def __init__(
      self, 
      save_path: str, 
      dir_output_name: str, 
      folders_to_ignore: list[str], 
      metadata_file_names: list[str]
      ):
      self._assert_directory_exists(save_path, dir_output_name)
      self._folders_to_ignore = {item for item in folders_to_ignore}
      self._save_path = save_path
      self._progress_bar = tqdm(total=100)
      self._folders_sum = 0
      self._metadata_file_names = {item for item in metadata_file_names}
      self._temp_mapping = {}
      self._dir_output_name = dir_output_name
      self._progress_bar_percentage = 0
   

   def _assert_directory_exists(
      self, 
      save_path: str, 
      dir_output_name: str
      ):
      """Tests if given directory exists.

      :param save_path: path to save.
      :param dir_output_name: name of output to save. 
      """
      path = os.path.join(save_path, dir_output_name)
      if os.path.exists(path):
         raise DirectoryAlreadyExists(path) 


   def __call__(
      self,
      dirs_to_map: str
      ):
      """Scan, store and output a directory.

      Scan all directories necessary, saving important data on the
      object's state, lastly, it prints some data about the operation
      and saves the output.
      
      :param dirs_to_map: directories to map.
      """
      for index, array in enumerate(dirs_to_map):
         path = array[0]
         nested_folders = int(array[1])
         load_bar_start = index/len(dirs_to_map)*100
         load_bar_end = (index+1)/len(dirs_to_map)*100
         self._read_dir(path, nested_folders, load_bar=True, load_bar_start=load_bar_start, load_bar_end=load_bar_end)
      self._write_map()
      self._update_progress_bar(percentage=100, post_fix="Finished", close=True)
    
      print(f"{self._folders_sum} items were scanned.")
      output_path = os.path.join(self._save_path, self._dir_output_name)
      if os.path.exists(output_path):
         print(f"Output saved on path: {output_path}")
      else:
         print(f"No content to save.")


   def _update_progress_bar(
      self, 
      percentage: float, 
      post_fix: str, 
      close: bool=False
      ):
      """Updates progress bar state.

      :param percentage: percentage the process is in.
      :param post_fix: hint to what item is being currently processed. 
      :param close: represents if the progress bar needs to be closed.
      """
      val = float(f"{percentage:.2f}")
      self._progress_bar.n = val
      self._progress_bar_percentage = val
      
      self._progress_bar.refresh()
      self._progress_bar.set_postfix_str(post_fix)
      if close:
         self._progress_bar.close()


   def _read_dir(
      self,
      directory: str, 
      max_rec: int, 
      current_rec: int=0, 
      load_bar: bool=False, 
      load_bar_start: int=0, 
      load_bar_end: int=100
      ):
      """Recursively reads folders and update objects state.

      Recursively calls itself, entering all the necessary sub folders. 
      Check for files with specific names and reads them, updating the 
      mapping of content and the progress bar.

      :param directory: directory to scan.
      :param max_rec: max subfolders to scan.
      :param current_rec: current subfolder count.
      :param load_bar: if the load bar will be updated.
      :param load_bar_start: the start of the load bar percentage.
      :param load_bar_end: the end of the load bar percentage.
      """

      if ("*" in self._metadata_file_names and directory.endswith(".dmmp")) \
                                 or any(directory.endswith(f"{item}.dmmp") \
                                 for item in self._metadata_file_names):
         self._get_desc_data(directory)
      if current_rec>max_rec or not os.path.isdir(directory):
         return
      folders = os.listdir(directory)
      
      for index, folder in enumerate(folders):
         if folder not in self._folders_to_ignore:  
            self._read_dir(
               fr"{directory}\{folder}",
               max_rec=max_rec, 
               current_rec=current_rec+1)
            self._folders_sum = self._folders_sum+1
         if load_bar:
            percentage = index/len(folders)
            diff = load_bar_end-load_bar_start
            relative_percentage = (percentage*diff)+load_bar_start
            self._update_progress_bar(
               percentage=relative_percentage, 
               post_fix=folder)


   def read_triple_string(self, file):
      """Read the triple string of a file
      
      :param file: the file object
      """
      inside = False
      buffer = []

      for line in file:
         if not inside:
               if '"""' in line:
                  inside = True
                  after = line.split('"""', 1)[1]
                  if after.strip():
                     buffer.append(after)
         else:
               if '"""' in line:
                  before = line.split('"""', 1)[0]
                  if before.strip():
                     buffer.append(before)
                  break
               buffer.append(line)

      return "".join(buffer)
   

   def _get_desc_data(
      self,
      directory: str
      ):
      """Get data from a directory and store its data.

      Find the file, open it, read line by line and store in the
      internal mapping.

      :param directory: file path to read.
      """
      with open(directory, "r", encoding="utf-8") as f:
            id = f.readline().strip()
            folder_path = f.readline().strip()
            name = f.readline().strip()
            triple_string = self.read_triple_string(f)
            
            desc = triple_string
            
            if self._temp_mapping.get(id, False):
               raise DuplicateUniqueId(id)

            self._temp_mapping[id] = {            
               "name":name,
               "desc":desc,
               "folder":folder_path,
               "origin":"/".join(directory.split("\\"))
            }


   def _write_map(self):
      """Write the mapping to a directory.

      Get all the data saved on the internal mapping and write it to a 
      tree of separate files in a specific directory.
      """
      self._assert_directory_exists(self._save_path, self._dir_output_name)

      path = os.path.join(self._save_path, self._dir_output_name)
      
      for id, object in self._temp_mapping.items():
         temp_path = os.path.join(path, object.get("folder"))
         os.makedirs(temp_path, exist_ok=True)
         id_to_change = re.findall(r"\[\[(.*?)\]\]", object.get("desc"))
         desc = object.get("desc")

         for id in id_to_change:
            desc = desc.replace(id, self._get_link(id))

         file_path = os.path.join(temp_path, f"{object.get("name")}.md")
         
         if os.path.exists(file_path):
            raise DuplicateOutputFile(file_path)
         
         with open(file_path, "w", encoding="utf-8") as f1:
            f1.write(f"{desc}\n\nid: {id}\norigin: {object.get("origin")}")


   def _get_link(
      self, 
      id
      ):
      """Get markdown friendly link from an id.

      Finds the output path of the item and use it to build a mardown 
      friendly link that can redirect to that path when clicked.

      :param id: id of the item.
      """
      obj = self._temp_mapping.get(id.strip())
      if obj:
         return f"{obj.get("folder")}/{obj.get("name")}|{obj.get("name")}"
      else:
         raise NoRelationshipFound(id)


class DirectoryAlreadyExists(Exception):
   def __init__(
      self, 
      path
      ):
      super().__init__(f"Output directory '{path}' already exists, cannot \
overwrite it")


class MissingArgumentException(Exception):
   def __init__(
      self, 
      arg
      ):
      super().__init__(f"Missing argument '--{arg.replace("_", "-")}'")


class DuplicateUniqueId(Exception):
   def __init__(
      self, 
      id
      ):
      super().__init__(f"Duplicate unique id: '{id}'")

      
class DuplicateOutputFile(Exception):
   def __init__(
      self, 
      path
      ):
      super().__init__(f"Two items with the same path and name: '{path}'")

      
class NoRelationshipFound(Exception):
   def __init__(
      self, 
      id
      ):
      super().__init__(f"Couldn't find relationship for id '{id}'")


import sys
from pathlib import Path
import pytest

new_lib_path = str(Path(__file__).resolve().parents[1])

sys.path.append(new_lib_path)

from src.dmmp.exceptions import DirectoryAlreadyExists
from src.dmmp.mapper import Mapper
import os
import shutil

mapper = Mapper("", "", ["desc"], ["*"])


def test_get_link():
   mapper._temp_mapping = {
      "id":{
         "folder":"/path/here",
         "name":"test"
      }
   }
   assert(mapper._get_link("id") == "/path/here/test|test")
   mapper._temp_mapping = {}

def test_get_desc_data():
   
   mapper._get_desc_data(os.path.join(
         "tests", 
         "test_content", 
         "desc.dmmp"))
   
   assert(
      mapper._temp_mapping == 
         {
            "123":{
               "name":"data",
               "desc":"data\ntest\n\n\ntest\n",
               "folder":"data/",
               "origin":"tests/test_content/desc.dmmp"
            }
         }
      )

def test_update_progress_bar():
   mapper._update_progress_bar(50, "Task 5")   
   assert(mapper._progress_bar.postfix == "Task 5")
   assert(mapper._progress_bar_percentage == 50)


mapper1_path = os.path.join(os.getcwd(), "tests", "test_content")
mapper1 = Mapper(
   mapper1_path,
   "temp",
   ["desc"],
   ["*"]
)

def test_write_map_and_assert_dir_exists():
   mapper1._temp_mapping = {
      "123":{
         "name":"data",
         "desc":"data",
         "folder":"data/",
         "origin":"tests/test_content/desc.dmmp"
      }
   }
   mapper1._write_map()
   
   with pytest.raises(DirectoryAlreadyExists):
      mapper1._write_map()

   with pytest.raises(DirectoryAlreadyExists):
      mapper1._assert_directory_exists(mapper1_path, "temp")
   
   shutil.rmtree(os.path.join(mapper1_path, "temp"))
   mapper1._assert_directory_exists(mapper1_path, "temp")



from dotenv import load_dotenv
import argparse
from .mapper import Mapper
from .args import get_args, validate_args, write_args_to_toml


if __name__ == "__main__":

   load_dotenv()  

   parser = argparse.ArgumentParser(
      prog="DMM",
      description="Directory Metadata Mapper"
   )
   add_arg = parser.add_argument
   add_arg('-c', '--config-file', 
            required=False, 
            help=".toml file to get configuration from ") 
   
   add_arg('-d', '--dirs-to-map', 
            required=False, 
            help="Directory to scan + number of nested folders to iterate", 
            nargs=2, 
            action="append")
   
   add_arg('-i', '--folders-ignore', 
            required=False, 
            help="List of folders to ignore", 
            nargs="+")
   
   add_arg('-s', '--save-path', 
            required=False, 
            help="Path to save output")
   
   add_arg('-n', '--name-output', 
            required=False, 
            help="Name of the output directory")
   
   add_arg('-m', '--metadata-file-names', 
            required=False, 
            help="File names containing the directory metadata (no extension)",
            nargs="+")
   
   args = parser.parse_args()
   
   formatted_args = get_args(**vars(args))

   required_args = [
      "dirs_to_map",
      "folders_ignore",
      "save_path",
      "name_output",
      "metadata_file_names",
      ]


   validate_args(formatted_args, required_args=required_args)

   mapper = Mapper(formatted_args.get("save_path"),
                  formatted_args.get("name_output"), 
                  formatted_args.get("folders_ignore"), 
                  formatted_args.get("metadata_file_names"))
   mapper(formatted_args.get("dirs_to_map"))
   
   write_args_to_toml(formatted_args, formatted_args["config_file"])